package stamp.peripheral.web;
import stamp.core.*;
import stamp.errors.*;
import stamp.util.text.*;

/**
 * This class provides access to the Pink webserver module.
 * There are methods to read/write Pink variables, read status and post register,
 * check several conditions, commands to send UDP and email messages.
 *
 * History
 * V1.4 April 13, 2006 Added writeBinVar() and readBinVar(). Bugfixed
 * V1.3 January 7, 2006 Added isBusy().
 * V1.2 November 14, 2005 Added '!','X' to sendEmail() to force 'Missformed message' error on Pink.
 * V1.1 November 9, 2005 Updated read and write methods, added clear method.
 * V1.0 November 3, 2005 Initial release (not tested).
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.0 November 3, 2005
 */
public class Pink {

  //special read/write registers (in order of memory map)
  public static final int Nb_varBM = 100; //content of udp message (both send and receive)
  public static final int Nb_varBI = 101; //IP destination address for udp message (send only)
  public static final int Nb_varEV = 102; //Email smtp server register
  public static final int Nb_varEC = 103; //Email Content register
  public static final int Nb_varES = 104; //Email Subject register
  public static final int Nb_varEF = 105; //Email From register
  public static final int Nb_varET = 106; //Email To register
  public static final int Nb_varST = 107; //status register, read only
  public static final int Nb_varSV = 108; //post register, read only

  private static final int CLS = 0;

  private int id;
  private Uart tx;
  private Uart rx;
  private ScaledTimer16 t;

  /*
    Web Variable Range    Maximum Size   Address
    Nb_var00 to Nb_var19    64 bytes      0-19
    Nb_var20 to Nb_var99    64 bytes     20-99
    Nb_varBM                64 bytes       100
    Nb_varBI                64 bytes       101
    Nb_varEV                64 bytes       102
    Nb_varEC                64 bytes       103
    Nb_varES                64 bytes       104
    Nb_varEF                64 bytes       105
    Nb_varET                64 bytes       106
    Nb_varST                 1 byte        107     //read only
    Nb_varSV                 1 byte        108     //read only
  */

  private static final char[] codes = new char[]{'B','M',  //100
                                                 'B','I',  //101
                                                 'E','V',  //102
                                                 'E','C',  //103
                                                 'E','S',  //104
                                                 'E','F',  //105
                                                 'E','T',  //106
                                                 'S','T',  //107 read only
                                                 'S','V'}; //108 read only

  private int status; //latest status
                      //b0 set if network link established
                      //b1 set if variable has been updated from webpage
                      //b2 set if Pink ready to send email
                      //b3 set if last email transmission had error
                      //b4 set if email has been sent
                      //b5 set if udp message received
                      //b6 set if Pink busy (b5-b1 are invalid then)
                      //b7 not used

  /**
   * Constructor for Pink object
   *
   * @param id Identifies Pink webserver (use default '0')
   * @param tx Transmitter used to send commands and data to Pink
   * @param rx Receiver uaed to receive status and data from Pink
   * @param t Timer to timeout receiver
   */
  public Pink(int id, Uart tx, Uart rx, ScaledTimer16 t) {
    this.id = id;
    this.tx = tx;
    this.rx = rx;
    this.t = t;
  }

  /*
    The Pink Webserver follows standard AppMod structure and is formatted
    as follows
           !NB0<CMD><y>
    where
           !      Preface character alerts units on the bus that the message is an AppMod command
           NB     Pink Webserver Prefix (2 byte address denoting the type of AppMod addressed)
           0      Unit ID (Pink module to be addresses by this command, 0 is the default)
           <CMD> - Command, i.e. W = Write Variable, R = Read Variable, etc.
           <y>   - Command or configuration parameter

    To write:
              send !!','N','B',ID,'W',<xx>,':',<definition>,CLS
    where
             <xx> - Variable reference number from 00 to 99, or Email or UDP codes (see Sections 4 and 5).
                    For example, when defining Nb_var20, use 20.
               :  - Separator between the definition and the command.
     <definition> - Definition of variable
              CLS - This lets the Pink know where the definition ends. This is necessary
                    because it is possible to store less data than the variable is capable
                    of holding. The CLS character is used to let the Pink know that it has
                    reached the end of the data for that variable.
    To read:
              send '!','N','B',ID,'R',<xx>
    where
             <xx> - Variable reference number from 00 to 99. For example,
                    when reading Nb_var05, use 05.
  */

  /**
   * Clear range of Pink variables.
   *
   * @param addr Address (0-106) of first Pink variable to clear.
   * @param num Number of variables (1-107) to clear.
   */
  public void clearVar(int addr, int num) {
    if ((addr<0) || (addr>106) || (num<1)) return; //do not write to PINK with invalid parameters
    while ((addr<107) && (num>0)) {
      sendCommand(addr++,true);
      tx.sendByte(':');
      tx.sendByte(CLS); //inform PINK this write command is completed
      num--;
    }
  }

  /**
   * Write value to Pink variable.
   *
   * @param addr Address (0-106) of Pink variable.
   * @param str String holding value to write.
   */
  public void writeVar(int addr, String str) {
    writeVar(addr,str.toCharArray(),str.length(),0);
  }

  /**
   * Write binary value to Pink variable.
   * If the number of bytes to write is 64 or more, this method writes 64 bytes to the
   * addressed variable and the remaining bytes to the next variable(s).
   * A '\' character is written as '\','\'
   * A binary 0 is written as '\','0'
   * If address reaches the value 107, any remaining bytes are not written.
   * No bytes are written if any of the parameters is invalid.
   *
   * @param addr Address (0-106) of Pink variable.
   * @param buf Array holding value to write.
   * @param num Number of bytes to write.
   * @param offset Start index in buf
   */
  public void writeBinVar(int addr, char[] buf, int num, int offset) {
    int i,c;
    if ((addr<0) || (addr>106) || (num<0) || (offset<0)) return; //do not write to PINK with invalid parameters
    int n = buf.length - offset; //maximum number of bytes to write
    if (n<0) return;             //invalid offset
    if (n<num) num = n;          //number of bytes to write
    if (num==0) {
      clearVar(addr,1);
      return;
    }
    while ((num>0) && (addr<107)) {
      sendCommand(addr,true);
      tx.sendByte(':');
      i=0;
      while ((num>0) && (i<64)) {
        num--;
        c = buf[offset++];
        if ((addr == 106) && (i==63)) break; //make sure the last byte of the last variable is CLS
        switch (c) {
          case '\\': tx.sendByte('\\'); // '\' send as '\','\'
                     tx.sendByte('\\');
                     i += 2;
                     break;
          case 0:    tx.sendByte('\\'); // 0 send as '\','0'
                     tx.sendByte('0');
                     i += 2;
                     break;
          default:   tx.sendByte(c);
                     i++;
        }
      }
      tx.sendByte(CLS); //inform PINK this write command is completed
      addr++;           //write to next PINK variable
    }
  }

  /**
   * Write value to Pink variable.
   * If the number of bytes to write is 64 or more, this method writes 64 bytes to the
   * addressed variable and the remaining bytes to the next variable(s).
   * If any of the bytes to write has the value CLS (binary 0) then the remaining bytes
   * are also written to the next variable(s).
   * If address reaches the value 107, any remaining bytes are not written.
   * No bytes are written if any of the parameters is invalid.
   *
   * @param addr Address (0-106) of Pink variable.
   * @param buf Array holding value to write.
   * @param num Number of bytes to write.
   * @param offset Start index in buf
   */
  public void writeVar(int addr, char[] buf, int num, int offset) {
    int i,c;
    if ((addr<0) || (addr>106) || (num<0) || (offset<0)) return; //do not write to PINK with invalid parameters
    int n = buf.length - offset; //maximum number of bytes to write
    if (n<0) return;             //invalid offset
    if (n<num) num = n;          //number of bytes to write
    if (num==0) {
      clearVar(addr,1);
      return;
    }
    while ((num>0) && (addr<107)) {
      sendCommand(addr,true);
      tx.sendByte(':');
      i=0;
      while ((num>0) && (i<64)) {
        num--;
        c = buf[offset++];
        if ((addr == 106) && (i==63)) break; //make sure the last byte of the last variable is CLS
        if (c != CLS) {
          tx.sendByte(c);
          i++;
        }
        else break;     //end of variable value, write remaining bytes to next variable
      }
      tx.sendByte(CLS); //inform PINK this write command is completed
      addr++;           //write to next PINK variable
    }
  }

  /**
   * Read binary value from Pink variable.
   * This method reads bytes from a variable until a CLS character is encountered,
   * or until the buffer is filled, whichever comes first.
   * A received '\' character is treated as an escape sequence.
   * The sequence '\','\' is treated as the character '\'
   * The sequence '\','0' is treated as a binary 0
   * A '\' followed by anything else is received as 2 bytes
   * No closing zero is appended.
   *
   * @param addr Address (0-106) of Pink variable.
   * @param buf Array to hold value.
   * @param num Number of bytes to read (1-32767, use 32767 to ensure read until buffer filled).
   * @param offset Start index in buf
   * @return Number of bytes read or errorcode in case of timeout or invalid parameter.
   */
  public int readBinVar(int addr, char[] buf, int num, int offset) {
    int i,c,prev;
    if ((addr<0) || (addr>106) || (num<0) || (offset<0)) return err.INVALID; //do not write to PINK with invalid parameters
    int n = buf.length - offset - 1;          //maximum number of bytes to read, last byte is cleared
    if (n<num) num = n;                       //number of bytes to read
    if ((n<0) || (num<1)) return err.INVALID; //invalid offset, no bytes read
    buf[buf.length-1] = 0;
    if (n==0) return 0;                       //return empty asciiz string
    sendCommand(addr,false);
    t.mark();
    i=0; c=-1; prev=-1;
    while ((c!=CLS) && !t.timeout()) { //keep reading until CLS, even if not stored in buf
      if (rx.byteAvailable()) {
        t.mark();
        prev = c;
        c = rx.receiveByte();
        if (i<n) {
          buf[offset+i] = (char)c; //only write to buf until position is buf.length-1
          if (i>0) {
            if (prev == '\\') {
              switch (c) {
                case '\\': i--;                // '\','\' received as '\'
                           c = -1;
                           break;
                case '0' : i--;
                           buf[offset+i] = 0;  // '\','0' received as 0
                           break;
                default:
              }
            }
          }
        }
        i++;
      }
    }
    return (c==CLS) ? i-1 : err.TIME_OUT;
  }

  /**
   * Read value from Pink variable.
   * This method reads bytes from a variable until a CLS character is encountered,
   * or until the buffer is filled, whichever comes first. A closing 0 will be
   * appended to the bytes that are read.
   *
   * @param addr Address (0-106) of Pink variable.
   * @param buf Array to hold value.
   * @param num Number of bytes to read (1-32767, use 32767 to ensure read until buffer filled).
   * @param offset Start index in buf
   * @return Number of bytes read (exclusive closing 0) or errorcode in case of timeout or invalid parameter.
   */
  public int readVar(int addr, char[] buf, int num, int offset) {
    int i,c;
    if ((addr<0) || (addr>106) || (num<0) || (offset<0)) return err.INVALID; //do not write to PINK with invalid parameters
    int n = buf.length - offset - 1;          //maximum number of bytes to read, last byte is cleared
    if (n<num) num = n;                       //number of bytes to read
    if ((n<0) || (num<1)) return err.INVALID; //invalid offset, no bytes read
    buf[buf.length-1] = 0;
    if (n==0) return 0;                       //return empty asciiz string
    sendCommand(addr,false);
    t.mark();
    i=0; c=-1;
    while ((c!=CLS) && !t.timeout()) { //keep reading until CLS, even if not stored in buf
      if (rx.byteAvailable()) {
        t.mark();
        c = rx.receiveByte();
        if (i<n) buf[offset+i] = (char)c; //only write to buf until position is buf.length-1
        i++;
      }
    }
    return (c==CLS) ? i-1 : err.TIME_OUT;
  }

  /**
   * Send email message.
   * Nb_varET, Nb_varEF, Nb_varES, Nb_varEC and Nb_varEV must contain correct values.
   */
  public void sendEmail() {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte('S');
    tx.sendByte('M');
    tx.sendByte('!'); //force Missformed message
    tx.sendByte('X');
  }

  /**
   * Send UDP message.
   * Nb_varBI and Nb_varBM must contain correct values.
   */
  public void sendUdp() {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte('S');
    tx.sendByte('B');
  }

  /**
   * Get status register.
   * Also internally stored for status methods.
   *
   * @return status or errorcode
   */
  public int getStatus() {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte('S');
    tx.sendByte('T');
    t.mark();
    while (!t.timeout()) {
      if (rx.byteAvailable()) {
        status = rx.receiveByte();
        return status;
      }
    }
    return err.TIME_OUT;
  }

  /**
   * Get post register.
   *
   * @return post or errorcode
   */
  public int getPost() {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte('S');
    tx.sendByte('V');
    t.mark();
    while (!t.timeout()) {
      if (rx.byteAvailable()) {
        return rx.receiveByte();
      }
    }
    return err.TIME_OUT;
  }

  /**
   * Check if network connection is established.
   * Method getStatus() must have been called.
   *
   * @return True if network connection established, false otherwise.
   */
  public boolean isConnected() {
    return (status & 0x01) != 0;
  }

  /**
   * Check if variable is updated via webpage.
   * Method getStatus() must have been called.
   * Nb_varSV will hold the number of the last variable updated from a web page POST.
   * To read the value of this variable, use readVar(Nb_varSV,buf,offset).
   *
   * @return True if variable updated, false otherwise.
   */
  public boolean isUpdated() {
    return !isBusy() && ((status & 0x02) != 0);
  }

  /**
   * Check if Pink ready to send email.
   * Method getStatus() must have been called.
   * Nb_varET, Nb_varEF, Nb_varES, Nb_varEC and Nb_varEV must contain correct values.
   *
   * @return True if Pink ready to send email, false otherwise.
   */
  public boolean isReady() {
    return !isBusy() && ((status & 0x04) == 0);
  }

  /**
   * Check if last email transmission had error.
   * Method getStatus() must have been called.
   *
   * @return True if email transmission had error, false otherwise.
   */
  public boolean isError() {
    return !isBusy() && ((status & 0x08) != 0);
  }

  /**
   * Check if email has been sent.
   * Method getStatus() must have been called.
   *
   * @return True if email has been sent, false otherwise.
   */
  public boolean isSent() {
    return !isBusy() && ((status & 0x10) != 0);
  }

  /**
   * Check if there is a new udp message.
   * Method getStatus() must have been called.
   *
   * @return True if new udp message received, false otherwise.
   */
  public boolean isUdp() {
    return !isBusy() && ((status & 0x20) != 0);
  }

  /**
   * Check if Pink is busy.
   * Method getStatus() must have been called.
   * If Pink is busy, it will only accept getStatus() command.
   *
   * @return True if Pink is busy, false otherwise.
   */
  public boolean isBusy() {
    return (status & 0x40) != 0;
  }

  /**
   * Send command to Pink webserver.
   *
   * @param addr Address (0-106) of Pink variable.
   * @param write True for write command, false for read command
   */
  private void sendCommand(int addr, boolean write) {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte(write ? 'W' : 'R');
    if (addr < 100) {
      tx.sendByte((addr/10)+'0');
      tx.sendByte((addr%10)+'0');
    }
    else {
      tx.sendByte(codes[(addr-100)<<1]);
      tx.sendByte(codes[((addr-100)<<1)+1]);
    }
  }

}